//
//  Visopsys
//  Copyright (C) 1998-2005 J. Andrew McLaughlin
// 
//  This program is free software; you can redistribute it and/or modify it
//  under the terms of the GNU General Public License as published by the Free
//  Software Foundation; either version 2 of the License, or (at your option)
//  any later version.
// 
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
//  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
//  for more details.
//  
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  kernelPS2MouseDriver.c
//

// Driver for PS2 meeses.

#include "kernelDriver.h" // Contains my prototypes
#include "kernelMouse.h"
#include "kernelGraphic.h"
#include "kernelMalloc.h"
#include "kernelProcessorX86.h"
#include <string.h>


static unsigned char inPort60(void)
{
  // Input a value from the keyboard controller's data port, after checking
  // to make sure that there's some data there for us

  unsigned char data = 0;

  while (!(data & 0x01))
    kernelProcessorInPort8(0x64, data);

  kernelProcessorInPort8(0x60, data);
  return (data);
}


static inline void waitControllerReady(void)
{
  // Wait for the controller to be ready
  unsigned char data = 0x02;
  while (data & 0x02)
    kernelProcessorInPort8(0x64, data);
}


static void outPort60(unsigned char value)
{
  // Output a value to the keyboard controller's data port, after checking
  // to make sure it's ready for the data

  unsigned char data = value;
  
  waitControllerReady();
  kernelProcessorOutPort8(0x60, data);

  return;
}


static void outPort64(unsigned char value)
{
  // Output a value to the keyboard controller's command port, after checking
  // to make sure it's ready for the command

  unsigned char data = value;
  
  waitControllerReady();
  kernelProcessorOutPort8(0x64, data);

  return;
}


static unsigned char getMouseData(void)
{
  // Input a value from the keyboard controller's data port, after checking
  // to make sure that there's some mouse data there for us

  unsigned char data = 0;

  while ((data & 0x21) != 0x21)
    kernelProcessorInPort8(0x64, data);

  kernelProcessorInPort8(0x60, data);
  return (data);
}


static void driverReadData(void)
{
  // This gets called whenever there is a mouse interrupt

  static volatile int button1, button2, button3;
  unsigned char byte1 = 0, byte2 = 0, byte3 = 0;
  int xChange, yChange;

  // Disable keyboard output here, because our data reads are not atomic
  outPort64(0xAD);

  // The first byte contains button information and sign information
  // for the next two bytes
  byte1 = getMouseData();

  // The change in X position
  byte2 = getMouseData();

  // The change in Y position
  byte3 = getMouseData();
  
  // Re-enable keyboard output
  outPort64(0xAE);

  if ((byte1 & 0x01) != button1)
    kernelMouseButtonChange(1, button1 = (byte1 & 0x01));
  else if ((byte1 & 0x04) != button2)
    kernelMouseButtonChange(2, button2 = (byte1 & 0x04));
  else if ((byte1 & 0x02) != button3)
    kernelMouseButtonChange(3, button3 = (byte1 & 0x02));

  else
    {
      // Sign them
      if (byte1 & 0x10)
	xChange = (int) ((256 - byte2) * -1);
      else
	xChange = (int) byte2;

      if (byte1 & 0x20)
	yChange = (int) (256 - byte3);
      else
	yChange = (int) (byte3 * -1);

      kernelMouseMove(xChange, yChange);
    }

  return;
}


static int driverDetect(void *driver)
{
  // This routine is used to detect and initialize each device, as well as
  // registering each one with any higher-level interfaces.  Also talks to
  // the keyboard controller a little bit to initialize the mouse

  int status = 0;
  kernelDevice *device = NULL;
  int interrupts = 0;
  unsigned char response = 0;
  unsigned char deviceId = 0;

  // Do the hardware initialization.

  // Disable keyboard output here, because our data reads are not atomic
  outPort64(0xAD);

  kernelProcessorSuspendInts(interrupts);

  // Send reset command
  outPort64(0xD4);
  outPort60(0xFF);

  // Read the ack 0xFA
  response = inPort60();
  if (response != 0xFA)
    goto exit;

  // Should be 'self test passed' 0xAA
  response = inPort60();
  if (response != 0xAA)
    goto exit;

  // Get the device ID.  0x00 for normal PS/2 mouse
  deviceId = inPort60();
  if (deviceId != 0)
    goto exit;

  // Set scaling to 2:1
  outPort64(0xD4);
  outPort60(0xE7);

  // Read the ack
  response = inPort60();
  if (response != 0xFA)
    goto exit;

  // Tell the controller to issue mouse interrupts
  outPort64(0x20);
  response = inPort60();
  response |= 0x02;
  outPort64(0x60);
  outPort60(response);

  // Enable data reporting (stream mode)
  outPort64(0xD4);
  outPort60(0xF4);

  // Read the ack
  response = inPort60();
  if (response != 0xFA)
    goto exit;

  // Allocate memory for the device
  device = kernelMalloc(sizeof(kernelDevice) + sizeof(kernelMouse));
  if (device == NULL)
    goto exit;

  device->class = kernelDeviceGetClass(DEVICECLASS_MOUSE);
  device->subClass = kernelDeviceGetClass(DEVICESUBCLASS_MOUSE_PS2);
  device->driver = driver;
  device->dev = ((void *) device + sizeof(kernelDevice));

  // Initialize mouse operations
  status = kernelMouseInitialize(device);
  if (status < 0)
    {
      kernelFree(device);
      goto exit;
    }

  // Add the device
  status = kernelDeviceAdd(NULL, device);
  if (status < 0)
    {
      kernelFree(device);
      goto exit;
    }

exit:
  kernelProcessorRestoreInts(interrupts);

  // Re-enable keyboard output
  outPort64(0xAE);

  return (status = 0);
}


static kernelMouseOps mouseOps = {
  driverReadData
};


/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
//
// Below here, the functions are exported for external use
//
/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////


void kernelPS2MouseDriverRegister(void *driverData)
{
   // Device driver registration.

  kernelDriver *driver = (kernelDriver *) driverData;

  driver->driverDetect = driverDetect;
  driver->ops = &mouseOps;

  return;
}
